#include "control.h"
#include "hogwash.h"
#include "../snort-1.8.1-RELEASE/snort.h"
#include <libnet.h>
#include <sys/stat.h>

/**************************************************
* Read the next line from the file into Buff
**************************************************/
int chogwash_get_next_line(int fd, char* Buff, int BuffLen){
	int		count;
	char	c;
	int		num_read;
	char*	this;
	
	num_read=0;
	this=Buff;
	while ((count=read(fd, &c, 1))!=-1){
		if (count==0) break;
		*this=c;
		if (c!=0x20){
		 this++;
		 num_read++;
		} 
		if (c==0x0A) break;
		if (num_read==BuffLen) break;
	}
	this--;
	*this=0;
	
	return num_read;
	
}

/**************************************************
* Get the next field out of the string
**************************************************/
int chogwash_get_next_field(char* InBuff, char StartChar, char EndChar, char* OutBuff, int OutBuffLen){
	int		count;
	char	c;
	int		num_read;
	char	*start, *end;
	
	num_read=0;
	start=InBuff;

	/*seek to the start char*/
	if (StartChar!=0){
		while (*start!=0x0 && *start!=StartChar) start++;
		if (*start==0x0) return 0;
		start++;
	}
	
	/*seek to the ending char*/
	end=start+1;
	while (*end!=0x0 && *end!=EndChar){
		num_read++;
		if (num_read==OutBuffLen) break;
		end++;
	}	
	num_read++;
	memcpy(OutBuff, start, num_read);
	
	return num_read;
	
}

/************************************************
* Read in the contents of the config file
************************************************/
int chogwash_read_config(char* config_file){
	int 	fd;
	char	LineBuff[1024];
	char	FieldBuff[1024];
	char	IntBuff[9];
	int		count;
	int		i;
	int		keynum;
	

#ifdef DEBUG
	printf("Reading in control config file\n");
#endif
	
	fd = open(config_file, O_RDONLY);
	if (fd==0){
		printf("Couldn't open config file\n");
		return FALSE;
	}
	
	while (chogwash_get_next_line(fd, LineBuff, 1024)){
		if (LineBuff[0]=='#') continue;
		if (LineBuff[0]==0x0A) continue;
		if (LineBuff[0]==0x0) continue;
		
		if (strncasecmp(LineBuff, "key", 3)==0){
			/*read in a twofish key*/
			bzero(FieldBuff,1024);
			count=chogwash_get_next_field(LineBuff, '(',',',FieldBuff, 1024);	
			keynum=atoi(FieldBuff);
			printf("Starting to add key number %i\n",keynum);
			if (keynum>=MAX_KEYS) continue;
			bzero(FieldBuff,1024);
			count=chogwash_get_next_field(LineBuff, ',',')',FieldBuff, 1024);	
			if (count!=((KeySize/32)*8)){
				printf("Wrong key size was %i characters expected %i characters\n",count,((KeySize/32)*8));
				continue;
			}	
			for (i=0;i<(KeySize/32);i++){
				bzero(&IntBuff[0],9);
				memcpy(IntBuff, FieldBuff+(i*8),8);
				IntBuff[9]=0x0;
				Keys[keynum].ki.key32[i]=strtol(IntBuff, NULL, 16); 
			}
			reKey(&Keys[keynum].ki);
			Keys[keynum].valid=TRUE;
			bzero(FieldBuff,1024);
			count=chogwash_get_next_field(LineBuff, ')',0,FieldBuff, 1024);	
			printf("Added key number %i\n",keynum);
		}else{
			printf("Unrecognized Keyword %s\n",LineBuff);
		}
	}
	
	close(fd);
	
	return TRUE;
}

/****************************************************
* Create a new file buffer
****************************************************/
int create_file_buffer(char* fname, unsigned int BuffID){
	int i;
	char*	startver;
	
	for (i=0;i<MAX_BUFFERS;i++){
		if (!Buffers[i].InUse) break;
	}
	
	if (i==MAX_BUFFERS) return FALSE;
	
	printf("Adding into Buffer %i\n",i);
	
	/*see if the buffer has a version*/
	startver=strstr(fname, ".v");
	if (!startver){
		strncpy(Buffers[i].FName, fname, 1024);
		Buffers[i].Version=-1;
	}else{
		memcpy(Buffers[i].FName, fname, startver-fname);
		Buffers[i].Version=atoi(startver+2);
	}
	
	Buffers[i].BuffID=BuffID;
	Buffers[i].BuffType = BUFFER_TYPE_SHARED_FILE;
	Buffers[i].DestroyTime=0;
	Buffers[i].InUse=TRUE;
	
	printf("FileName is %s\n",Buffers[i].FName);
	printf("Verion is %i\n",Buffers[i].Version);
	
	return TRUE;
}

/***************************************************
* Parse the command line options for the control channel
***************************************************/
void control_cmd_line(int argc, char**argv){
	int i;
	
	for (i=0;i<argc;i++){
		if (strncasecmp("-k", argv[i],2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=='-')){
				printf("Syntax -k <control config file>\n");
				exit(1);
			}
			CV.UseControl=TRUE;
			snprintf(CV.ConfigFileName, 1024, argv[i+1]);
		}
	}
}

/****************************************************
* Set up everything to do control channel stuff
****************************************************/
void control_init(pcap_t* pad, int argc, char* argv[]){
	int i,j;
	
	printf("Initializing Control Channel\n");
	
	printf("Parsing Control args\n");
	bzero(&CV, sizeof(struct control_vars));
	control_cmd_line(argc, argv);
	
	/*create stock buffers*/
	bzero(Buffers, sizeof(XBuff) * MAX_BUFFERS);
	for (j=0;j<MAX_BUFFERS;j++){
		Buffers[j].Version=-1;
	}
	create_file_buffer(pv.config_file, STOCK_BUFFER_RULES);
	create_file_buffer(LastAlertFileName, STOCK_BUFFER_ALERT);
	create_file_buffer(LastLogFileName, STOCK_BUFFER_LOG);
	create_file_buffer(pv.config_file, STOCK_BUFFER_NEW_RULES);
	
	printf("Initializing keys\n");
	for (j=0;j<MAX_KEYS;j++){
		makeKey(&Keys[j].ki, DIR_ENCRYPT, KeySize, NULL);
		cipherInit(&Keys[j].ci, MODE_ECB, NULL);
		Keys[j].valid=FALSE;
	}	
	
	chogwash_read_config(CV.ConfigFileName);
}

/***********************************************
* Builds a packet to send back to the requester
*
* NewPacket must be malloc'd by the caller
************************************************/
int BuildReturnPacket(char* NewPacket, int* NewLen, char* OldPacket, int OldLen, char* Message, int MessageLen){
	unsigned char 	Transport;
	int				PacketLen;
	int				iplen;
	int				tcplen;

	/*determine transport layer*/
	Transport=*(unsigned char*)(OldPacket+23);
	iplen=((*(OldPacket+14)) & 0x0F)*4;

	libnet_build_ethernet(
		OldPacket+6,
		OldPacket,
		ETHERTYPE_IP,
		NULL,
		0,
		NewPacket);
	
	/*build the ip header*/
	switch (Transport){
		case 17:
			PacketLen=8 + MessageLen;
			break;
		case 6:
			break;
		case 1:
			break;
		default:
			printf("Dont' know how to build that kind of packet\n");
			return FALSE;
	}
	
	libnet_build_ip(
		PacketLen,
		0,
		555 * rand(),
		0,
		64,
		Transport,
		*((unsigned long*)(OldPacket +14 + 16)),
		*((unsigned long*)(OldPacket +14 + 12)),
		NULL,
		0,
		NewPacket + LIBNET_ETH_H);
	
	/*build transport layer*/
	switch (Transport){
	case 17:
		libnet_build_udp(
			htons(*((unsigned short*)(OldPacket+14+iplen+2))),
			htons(*((unsigned short*)(OldPacket+14+iplen))),
			Message,
			MessageLen,
			NewPacket + LIBNET_ETH_H + LIBNET_IP_H);
		libnet_do_checksum(NewPacket + ETH_H, IPPROTO_UDP, LIBNET_UDP_H+ MessageLen);	
		libnet_do_checksum(NewPacket + ETH_H, IPPROTO_IP, LIBNET_IP_H);
		*NewLen=14 + 20 + 8 + MessageLen;
		break;
	case 6:
	case 1:
	}
	
	return TRUE;
}

/*******************************************
* Read the contents of a file into a buffer
********************************************/
int ReadFile(char* fname, char* Buff, unsigned int offset, unsigned int len, int* flags){
	FILE*			fd;
	int				count;
	struct stat		st;
	unsigned int	read_len;
	
	printf("@@@@@@Requested %u bytes starting at %u\n",len, offset);
	
	*flags=BUFFER_FLAG_OK;
	if (stat(fname, &st)!=0){
		*flags=BUFFER_FLAG_NOT_FOUND;
		printf("BUFFER_FLAG_NOT_FOUND\n");
		printf("Error Number is %i\n",errno);
		printf("fname is %s\n",fname);
		switch (errno){
		case EBADF:
			printf("EBADF\n");
			break;
		case ENOENT:
			printf("ENOENT\n");
			break;	
		default:
			printf("No Idea\n");
		}
		return 0;
	}
	if (offset>st.st_size){
		*flags=BUFFER_FLAG_END_OF_BUFFER;
		printf("st_size %i\n",st.st_size);
		printf("offset  %i\n",offset);
		printf("BUFFER_FLAG_END_OF_BUFFER\n");
		return 0;
	}
	fd=fopen(fname, "r");
	if (!fd){
		*flags=BUFFER_FLAG_OPEN_ERROR;
		printf("BUFFER_FLAG_OPEN_ERROR\n");	
		return 0;
	}
	if (fseek(fd, offset, SEEK_SET)!=0){
		*flags=BUFFER_FLAG_SEEK_ERROR;
		printf("BUFFER_FLAG_SEEK_ERROR\n");
		return 0;
	} 
	read_len=len;
	if (offset+len > st.st_size){
		read_len=st.st_size-offset;
		printf("st_size %i\n",st.st_size);
		printf("offset  %i\n",offset);
		printf("changed length to %i\n",read_len);
	}
	count=fread(Buff, read_len, 1, fd);
	fclose(fd);
	
	printf("I read %s\n",Buff);
	printf("@@@@@@@Returning length %u\n",read_len);
	return read_len;
}

/*******************************************
* Write the contents of a buffer to a file
********************************************/
int WriteFile(char* fname, char* Buff, unsigned int offset, unsigned int len, int* flags){
	FILE*			fd;
	int				count;
	struct stat		st;
	
	printf("@@@@@@Writing file %s %u bytes starting at %u\n",fname, len, offset);
	
	*flags=BUFFER_FLAG_OK;
	if (stat(fname, &st)==-1){
		fd=fopen(fname, "w+");
	}else{
		fd=fopen(fname, "r+");
	}	
	if (!fd){
		*flags=BUFFER_FLAG_OPEN_ERROR;
		printf("BUFFER_FLAG_OPEN_ERROR\n");	
		return 0;
	}
	if (fseek(fd, offset, SEEK_SET)!=0){
		*flags=BUFFER_FLAG_SEEK_ERROR;
		printf("BUFFER_FLAG_SEEK_ERROR\n");
		return 0;
	} 
	count=fwrite(Buff, (int)len, 1, fd);
	fclose(fd);
	
	printf("I wrote %s\n",Buff);
	printf("@@@@@@@Returning length %u\n",len);
	return len;
}


/*****************************************
* Handle the ping request
******************************************/
int ProcessPing(PacketToProcess* ptp, char* Pay, int PayLen, int KeyNum){
	ControlPacketPing*		cpp;
	ControlPacketPingReply	cppr;
	int						count;
	char					Buff[65536];

	if (PayLen!=sizeof(ControlPacketPing)) return FALSE;
	
	cpp=(ControlPacketPing*)Pay;
	
#ifdef DEBUG1
	printf("Ping Packet:\n");
	printf("MagicToken1: %u\n",cpp->MagicToken1);
	printf("MagicToken2: %u\n",cpp->MagicToken2);
	printf("PacketType : %u\n",cpp->PacketType);
	printf("PacketID   : %u\n",cpp->PacketID);
	printf("Message    : %s\n",cpp->PingMessage);
#endif	
	
	cppr.MagicToken1=CONTROL_MAGIC_TOKEN;
	cppr.MagicToken2=CONTROL_MAGIC_TOKEN;
	cppr.MagicToken2=CONTROL_PACKET_PING_REPLY;
	cppr.PacketID=cpp->PacketID;
	sprintf(cppr.PingMessage, cpp->PingMessage);
	
	*(unsigned int*)Buff=cppr.MagicToken1;
	count=blockEncrypt(&Keys[KeyNum].ci, &Keys[KeyNum].ki, (char*)&cppr+4, (sizeof(ControlPacketPingReply)-4)*8, Buff+4);
	if (count!=(sizeof(ControlPacketPing)-4)*8){
		printf("Failed to encrypt returned %li\n",count);
	}
	
	return BuildReturnPacket((u_char*)ptp->Buffer, 
		&ptp->BuffLen, 
		ptp->pkt, 
		ptp->pkthdr->len, 
		Buff, 
		sizeof(ControlPacketPingReply)
	);
}


/*****************************************
* Handle the Buffer request
******************************************/
int ProcessBuff(PacketToProcess* ptp, char* Pay, int PayLen, int KeyNum){
	ControlPacketBuffRequest*		cpb;
	ControlPacketBuffReply			cpbr;
	int								count;
	char							Buff[65536];
	int								i;
	char							RealFileName[1024];

#ifdef DEBUG
	printf("About to Check Payload Length\n");
	printf("Payload len %i\n", PayLen);
	printf("Structure %i\n",sizeof(ControlPacketBuffRequest));
#endif
	if (PayLen!=sizeof(ControlPacketBuffRequest)) return FALSE;
	
	cpb=(ControlPacketBuffRequest*)Pay;
	
#ifdef DEBUG1
	printf("Buff Request Packet:\n");
	printf("MagicToken1: %u\n",cpb->MagicToken1);
	printf("MagicToken2: %u\n",cpb->MagicToken2);
	printf("PacketType : %u\n",cpb->PacketType);
	printf("PacketID   : %u\n",cpb->PacketID);
#endif	
	
	cpbr.MagicToken1=CONTROL_MAGIC_TOKEN;
	cpbr.MagicToken2=CONTROL_MAGIC_TOKEN;
	cpbr.PacketType=CONTROL_PACKET_BUFF_REPLY;
	cpbr.PacketID=cpb->PacketID;
	cpbr.BuffID=cpb->BuffID;
	cpbr.Offset=cpb->Offset;
	cpbr.Len=cpb->Len;
	cpbr.Flags=0;
	
	/*find the buffer*/
	for (i=0;i<MAX_BUFFERS; i++){
		if (Buffers[i].BuffID==cpbr.BuffID) break;
	}
	
	if (Buffers[i].Version==-1){
		sprintf(RealFileName, Buffers[i].FName);
	}else{
		sprintf(RealFileName, "%s.v%05i", Buffers[i].FName, Buffers[i].Version);
	}
	printf("Real file name is %s\n",RealFileName);
	
	if (i==MAX_BUFFERS){printf("Buffer ID not found\n");return 0;}
	if (cpbr.Len<=MAX_BUFFER_SEGMENT){
		cpbr.Len=ReadFile(RealFileName, cpbr.Data, cpbr.Offset, cpbr.Len, &cpbr.Flags);
	}
	
	*(unsigned int*)Buff=cpbr.MagicToken1;
	count=blockEncrypt(&Keys[KeyNum].ci, &Keys[KeyNum].ki, (char*)&cpbr+4, (sizeof(ControlPacketBuffReply)-4)*8, Buff+4);
	if (count!=(sizeof(ControlPacketBuffReply)-4)*8){
		printf("Failed to encrypt returned %li\n",count);
	}
	
	return BuildReturnPacket((u_char*)ptp->Buffer, 
		&ptp->BuffLen, 
		ptp->pkt, 
		ptp->pkthdr->len, 
		Buff, 
		sizeof(ControlPacketBuffReply)
	);
}


/*****************************************
* Handle the Buffer Status request
******************************************/
int ProcessStat(PacketToProcess* ptp, char* Pay, int PayLen, int KeyNum){
	ControlPacketStatRequest*		cps;
	ControlPacketStatReply			cpsr;
	int								count;
	char							Buff[65536];
	struct stat						st;
	int								i;

#ifdef DEBUG
	printf("About to Check Payload Length\n");
	printf("Payload len %i\n", PayLen);
	printf("Structure %i\n",sizeof(ControlPacketStatRequest));
#endif
	if (PayLen!=sizeof(ControlPacketStatRequest)) return FALSE;
	
	cps=(ControlPacketStatRequest*)Pay;
	
#ifdef DEBUG1
	printf("Stat Request Packet:\n");
	printf("MagicToken1: %u\n",cps->MagicToken1);
	printf("MagicToken2: %u\n",cps->MagicToken2);
	printf("PacketType : %u\n",cps->PacketType);
	printf("PacketID   : %u\n",cps->PacketID);
	printf("BufferID   : %u\n",cps->BuffID);
#endif	
	
	cpsr.MagicToken1=CONTROL_MAGIC_TOKEN;
	cpsr.MagicToken2=CONTROL_MAGIC_TOKEN;
	cpsr.PacketType=CONTROL_PACKET_STAT_REPLY;
	cpsr.PacketID=cps->PacketID;
	cpsr.BuffID=cps->BuffID;
	
	if (cps->BuffID!=0){
		/*find the buffer*/
		for (i=0;i<MAX_BUFFERS; i++){
			if (Buffers[i].InUse)
			if (Buffers[i].BuffID==cpsr.BuffID) break;
		}
	
		if (i==MAX_BUFFERS){printf("Couldn't find buffer\n");return 0;}
		if (stat(Buffers[i].FName, &st)!=0){
			printf("Couldn't stat file%i (Buff %u) *%s*\n",i,Buffers[i].BuffID, Buffers[i].FName);
		}
	
		/*Buffer stats*/
		cpsr.BuffType=BUFFER_TYPE_SHARED_FILE;
		cpsr.CreateTime=st.st_ctime;
		cpsr.ModifyTime=st.st_mtime;
		cpsr.TimeLeft=0;
		cpsr.Size=st.st_size;
		cpsr.BufferVersion=Buffers[i].Version;
	}
	
	/*snort stats*/
	cpsr.TCPCount=pc.tcp;
	cpsr.UDPCount=pc.udp;
	cpsr.ICMPCount=pc.icmp;
	cpsr.ARPCount=pc.arp;
	cpsr.IPV6Count=pc.ipv6;
	cpsr.IPXCount=pc.ipx;
	cpsr.OtherCount=pc.other;
	cpsr.AlertCount=pc.alert_pkts;
	cpsr.LogCount=pc.log_pkts;
	cpsr.DropCount=0;	/*Not yet implemented*/
		
	*(unsigned int*)Buff=cpsr.MagicToken1;
	count=blockEncrypt(&Keys[KeyNum].ci, &Keys[KeyNum].ki, (char*)&cpsr+4, (sizeof(ControlPacketStatReply)-4)*8, Buff+4);
	if (count!=(sizeof(ControlPacketStatReply)-4)*8){
		printf("Failed to encrypt returned %li\n",count);
	}
	
	return BuildReturnPacket((u_char*)ptp->Buffer, 
		&ptp->BuffLen, 
		ptp->pkt, 
		ptp->pkthdr->len, 
		Buff, 
		sizeof(ControlPacketStatReply)
	);
}

/*****************************************
* Handle the Buffer Status request
******************************************/
int ProcessSetBuff(PacketToProcess* ptp, char* Pay, int PayLen, int KeyNum){
	ControlPacketSetBuffer*			cpsb;
	ControlPacketSetBufferReply		cpsbr;
	int								count;
	char							Buff[65536];
	struct stat						st;
	int								i;
	char							RealFileName[1024];

#ifdef DEBUG
	printf("About to Check Payload Length\n");
	printf("Payload len %i\n", PayLen);
	printf("Structure %i\n",sizeof(ControlPacketSetBuffer));
#endif
	if (PayLen!=sizeof(ControlPacketSetBuffer)) return FALSE;
	
	cpsb=(ControlPacketSetBuffer*)Pay;
	
#ifdef DEBUG1
	printf("Set Buffer Packet:\n");
	printf("MagicToken1: %u\n",cpsb->MagicToken1);
	printf("MagicToken2: %u\n",cpsb->MagicToken2);
	printf("PacketType : %u\n",cpsb->PacketType);
	printf("PacketID   : %u\n",cpsb->PacketID);
	printf("BufferID   : %u\n",cpsb->BuffID);
#endif	

	/*find the buffer*/
	for (i=0;i<MAX_BUFFERS; i++){
		if (Buffers[i].InUse)
		if (Buffers[i].BuffID==cpsb->BuffID) break;
	}
	if (i==MAX_BUFFERS){printf("Couldn't find buffer\n");return 0;}
	
	if (Buffers[i].Version==-1){
		sprintf(RealFileName, Buffers[i].FName);
	}else{
		sprintf(RealFileName, "%s.v%05i", Buffers[i].FName, Buffers[i].Version);
	}
	printf("Real file name is %s\n",RealFileName);

	count=WriteFile(RealFileName, cpsb->Data, cpsb->Offset, cpsb->Len, &cpsb->Flags);
	
	cpsbr.MagicToken1=CONTROL_MAGIC_TOKEN;
	cpsbr.MagicToken2=CONTROL_MAGIC_TOKEN;
	cpsbr.PacketType=CONTROL_PACKET_SET_REPLY;
	cpsbr.Flags=cpsb->Flags;
	cpsbr.PacketID=cpsb->PacketID;
	cpsbr.Offset=cpsb->Offset;
	cpsbr.Len=cpsb->Len;

	*(unsigned int*)Buff=cpsbr.MagicToken1;
	count=blockEncrypt(&Keys[KeyNum].ci, &Keys[KeyNum].ki, (char*)&cpsbr+4, (sizeof(ControlPacketSetBufferReply)-4)*8, Buff+4);
	if (count!=(sizeof(ControlPacketSetBufferReply)-4)*8){
		printf("Failed to encrypt returned %li\n",count);
	}
	
	return BuildReturnPacket((u_char*)ptp->Buffer, 
		&ptp->BuffLen, 
		ptp->pkt, 
		ptp->pkthdr->len, 
		Buff, 
		sizeof(ControlPacketSetBufferReply)
	);
		
	return FALSE;
}

/*****************************************
* Handle the Command request
******************************************/
int ProcessCommand(PacketToProcess* ptp, char* Pay, int PayLen, int KeyNum){
	ControlPacketCommand*			cpc;
	ControlPacketCommandReply		cpcr;
	int								count;
	char							Buff[65536];
	struct stat						st;
	int								i;

#ifdef DEBUG
	printf("About to Check Payload Length\n");
	printf("Payload len %i\n", PayLen);
	printf("Structure %i\n",sizeof(ControlPacketCommand));
#endif
	if (PayLen!=sizeof(ControlPacketCommand)) return FALSE;
	
	cpc=(ControlPacketCommand*)Pay;
	
#ifdef DEBUG1
	printf("Command Packet:\n");
	printf("MagicToken1: %u\n",cpc->MagicToken1);
	printf("MagicToken2: %u\n",cpc->MagicToken2);
	printf("PacketType : %u\n",cpc->PacketType);
	printf("PacketID   : %u\n",cpc->PacketID);
	printf("BufferID   : %u\n",cpc->BuffID);
#endif	

	switch(cpc->CommandID){
	case COMMAND_RESTART:
		printf("Executing Restart Command\n");
		Restart=TRUE;
		kill(MainPID, SIGHUP);
		kill(Thread1PID, SIGHUP);
		kill(Thread2PID, SIGHUP);
		return FALSE;
	case COMMAND_SET_BUFFER_VERSION:
		printf("Set Buffer Version\n");
		/*find the buffer*/
		for (i=0;i<MAX_BUFFERS; i++){
			if (Buffers[i].InUse)
			if (Buffers[i].BuffID==cpc->BuffID) break;
		}
		if (i==MAX_BUFFERS){printf("Couldn't find buffer\n");cpcr.Flags=FALSE;break;}

		/*set the new version*/
		Buffers[i].Version=*((int*)cpc->Data);
		printf("Version set to %i\n",Buffers[i].Version);
		cpcr.Flags=TRUE;
		break;
	default:
		printf("Unknown Command\n");
		return FALSE;
	}

	cpcr.MagicToken1=CONTROL_MAGIC_TOKEN;
	cpcr.MagicToken2=CONTROL_MAGIC_TOKEN;
	cpcr.PacketType=CONTROL_PACKET_COMMAND_REPLY;
	cpcr.PacketID=cpc->PacketID;
	cpcr.BuffID=cpc->BuffID;
	cpcr.CommandID=cpc->CommandID;

	*(unsigned int*)Buff=cpcr.MagicToken1;
	count=blockEncrypt(&Keys[KeyNum].ci, &Keys[KeyNum].ki, (char*)&cpcr+4, (sizeof(ControlPacketCommandReply)-4)*8, Buff+4);
	if (count!=(sizeof(ControlPacketCommandReply)-4)*8){
		printf("Failed to encrypt returned %li\n",count);
	}
	
	return BuildReturnPacket((u_char*)ptp->Buffer, 
		&ptp->BuffLen, 
		ptp->pkt, 
		ptp->pkthdr->len, 
		Buff, 
		sizeof(ControlPacketCommandReply)
	);
		
	return FALSE;
}


/****************************************/
/*find the control payload in the packet*/
/****************************************/
u_char* control_find_payload(u_char* pkt, unsigned int len, unsigned int* payload_len){
	u_char*			payload;
	unsigned short*	v1;	
	unsigned char*	v2;
	int				iplen,tcplen;
	
	if (len<256){
#ifdef DEBUG	
		printf("Packet was too short\n");
#endif		
		return NULL;
	}
	
	/*assume an ethernet packet right now*/
	v1=(unsigned short*)(pkt+12);
	if (*v1!=8){
#ifdef DEBUG	
		printf("Error in Ethernet Header. Not and IP packet\n");
#endif		
		return NULL;
	}
	
	v2=(unsigned char*)(pkt+23);
	iplen=((*(pkt+14)) & 0x0F)*4;
	switch (*v2){
	case 17:
#ifdef DEBUG
		printf("Was a UDP packet\n");
#endif
		payload=pkt+14+iplen+8;
		*payload_len=len-(14+iplen+8);
		break;	
	case 6:
#ifdef DEBUG
		printf("Was a TCP packet\n");
#endif
		tcplen=((*(pkt+14+iplen+13)) & 0xF0)/8 * 4;
		payload=pkt+14+iplen+tcplen;
		*payload_len=len-(14+iplen+tcplen);
		break;		
	case 1:
#ifdef DEBUG
		printf("Was an ICMP packet\n");
#endif
		payload=pkt+14+iplen+4;
		*payload_len=len-(14+iplen+4);
		break;		
	default:
#ifdef DEBUG
		printf("Can't tunnel over this packe type");
#endif
		return NULL;	
	}
	
	if (*((unsigned int*)payload)!=CONTROL_MAGIC_TOKEN){
#ifdef DEBUG
		printf("Didn't have the magic token\n");
#endif	
		return NULL;
	}
		
	printf("**This is a control packet**\n");
	return payload;
}


/*********************************************************
* All packets hit here
*********************************************************/
int control_func(PacketToProcess* ptp){
	u_char*				pay;
	u_char				buff[4096];
	int					count;
	unsigned int		payload_len;
	ControlPacket*		cp;
	int					i;

#ifdef DEBUG	
	printf("Control got a packet\n");
#endif
	
	/*check for the magic token*/
	pay=control_find_payload(ptp->pkt, ptp->pkthdr->len, &payload_len);
	if (!pay){
#ifdef DEBUG	
		printf("Wasn't a control packet\n");
#endif		
		return RESPONSE_SEND_ON;
	}
	
#ifdef DEBUG
	printf("About to Decrypt the packet\n");
#endif	

	/*decrypt the packet*/
	cp=(ControlPacket*)buff;
	cp->MagicToken1=*((unsigned int*)pay);
	for (i=0;i<MAX_KEYS;i++){
		count=0;
		if (Keys[i].valid)
		count=blockDecrypt(&Keys[i].ci, &Keys[i].ki, pay+4, (payload_len-4)*8, buff+4);
		if (count<1) continue;
		if (cp->MagicToken2==CONTROL_MAGIC_TOKEN) break;
	}	
	
#ifdef DEBUG1
	printf("I'm using key number %i\n",i);
#endif
		
	if (i==MAX_KEYS) return RESPONSE_SEND_ON;	
		
#ifdef DEBUG
	printf("Handling packet type %u\n",cp->PacketType);
#endif
	
	/*handle the packet*/
	switch (cp->PacketType){
	case CONTROL_PACKET_PING:
		if (ProcessPing(ptp, buff, payload_len, i)){
			return RESPONSE_SEND_REPLY | RESPONSE_USE_BUFFER;
		}else{
			return RESPONSE_DROP;
		}	
	case CONTROL_PACKET_PING_REPLY:
		return RESPONSE_DROP;	
	case CONTROL_PACKET_BUFF_REQUEST:
		if (ProcessBuff(ptp, buff, payload_len, i)){
			return RESPONSE_SEND_REPLY | RESPONSE_USE_BUFFER;
		}else{
			return RESPONSE_DROP;
		}	
	case CONTROL_PACKET_BUFF_REPLY:
		return RESPONSE_DROP;	
	case CONTROL_PACKET_STAT_REQUEST:
		if (ProcessStat(ptp, buff, payload_len, i)){
			return RESPONSE_SEND_REPLY | RESPONSE_USE_BUFFER;
		}else{
			return RESPONSE_DROP;
		}
	case CONTROL_PACKET_STAT_REPLY:
		return RESPONSE_DROP;
	case CONTROL_PACKET_SET_BUFF:
		if (ProcessSetBuff(ptp, buff, payload_len, i)){
			return RESPONSE_SEND_REPLY | RESPONSE_USE_BUFFER;
		}else{
			return RESPONSE_DROP;
		}	
	case CONTROL_PACKET_SET_REPLY:
		return RESPONSE_DROP;	
	case CONTROL_PACKET_COMMAND:
		if (ProcessCommand(ptp, buff, payload_len, i)){
			return RESPONSE_SEND_REPLY | RESPONSE_USE_BUFFER;
		}else{
			return RESPONSE_DROP;
		}	
	case CONTROL_PACKET_COMMAND_REPLY:
		return RESPONSE_DROP;	
	
	default:
#ifdef DEBUG
		printf("I don't know how to handle PacketType %u\n",cp->PacketType);
#endif
		return RESPONSE_DROP;
	}

}


